class Node():
    def __init__(self, key, vp):
        self.key = key
        self.vp = vp
        self.left = None
        self.right = None


class AVLTree():
    """
    The `AVLTree()` object used by the `Ring` object to
    implement the consistent hashing ring. 
    """

    def __init__(self):
        self.node = None
        self.height = -1
        self.balance = 0


    def insert(self, key, vp):
        """
        The `insert()` method inserts a key/value pair into
        the AVL tree.

        The `key` parameter is the hexidecimal representation 
        of the CRC32 hash of the GhostDB node IP address.

        The `vp` parameter is the `VirtualPoint` being inserted
        into the tree.
        """

        node = Node(key, vp)

        # Tree is empty, initialize root
        if self.node == None:
            self.node = node
            self.node.left = AVLTree()
            self.node.right = AVLTree()
        # Insert into left subtree
        elif key < self.node.key:
            self.node.left.insert(key, vp)
        # Insert into right subtree
        elif key > self.node.key:
            self.node.right.insert(key, vp)
        
        # Rebalance if needed
        self.__rebalance()


    def remove(self, key):
        """
        The `remove()` method removes a key/value pair from the
        tree where the `value` is a `VirtualPoint` object.

        The `key` parameter is the hexidecimal representation 
        of the CRC32 hash of the GhostDB node IP address.
        """

        if self.node != None:
            # Found the node delete it
            if self.node.key == key:
                # Node is a leaf node - Just remove
                if not self.node.left.node and not self.node.right.node:
                    self.node = None
                # Node has only one subtree - the right subtree - replace root with that
                elif not self.node.left.node:
                    self.node = self.node.right.node
                # Node has only one subtree - the left subtree - replace root with that
                elif not self.node.right.node:
                    self.node = self.node.left.node
                else:
                    # Find successor as smallest node in the right subtree or
                    # predecessor as largest node in left subtree
                    successor = self.node.right.node
                    while successor and successor.left.node:
                        successor = successor.left.node

                    if successor:
                        self.node.key = successor.key
                        # Delete successor from the replaced node right subtree
                        self.node.right.remove(successor.key)

            # Remove from left subtree
            elif key < self.node.key:
                self.node.left.remove(key)
            # Remove from right subtree
            elif key > self.node.key:
                self.node.right.remove(key)
            
            # Rebalance if needed
            self.__rebalance()


    def minimum_pair(self):
        """
        the `minimum_pair()` method returns the key with the 
        smallest key value.
        """
        
        if self.node == None:
            return None
        
        current_node = self.node
        while current_node.left.node != None:
            current_node = current_node.left.node

        return (current_node.key, current_node.vp)


    def next_gte_pair(self, key):
        """
        The `next_gte_pair()` method retrieves the next GhostDB 
        node in the tree that follows the provided key in sorted
        order.

        The `key` parameter is the hexidecimal representation 
        of the CRC32 hash of the GhostDB node IP address.
        """

        node = self.__next_gte_node(self.node, key)

        if node == None:
            return None, None
        return (node.key, node.vp)


    def is_balanced(self):
        """
        The `is_balanced()` method returns `True` if the `AVLTree`
        instance is balanced, otherwise it returns `False`
        """
        
        if self.node is None:
            return True
        
        # Left subtree height
        lst = self.node.left.height
        # Right subtree height
        rst = self.node.right.height

        if (abs(lst - rst) <= 1) and self.node.left.is_balanced() is True and self.node.right.is_balanced() is True:
            return True

        return False


    def get_nodes(self):
        """
        The `get_nodes()` method returns a list of all `VirtualPoint`
        objects in the tree.
        """

        nodes = []

        if not self.node:
            return nodes
        
        nodes.extend(self.node.left.get_nodes())
        nodes.append(self.node.vp)
        nodes.extend(self.node.right.get_nodes())

        return nodes


    def inorder_traverse(self):
        """
        The `inorder_traverse()` method returns a list of all keys
        in the tree in in-order order
        """

        keys = []

        if not self.node:
            return keys
        
        keys.extend(self.node.left.inorder_traverse())
        keys.append(self.node.vp.index)
        keys.extend(self.node.right.inorder_traverse())

        return keys

    def preorder_traverse(self):
        """
        The `preorder_traverse()` method returns a list of all keys
        in the tree in pre-order order
        """
        
        keys = []

        if not self.node:
            return keys
        
        keys.append(self.node.vp.index)
        keys.extend(self.node.left.preorder_traverse())
        keys.extend(self.node.right.preorder_traverse())

        return keys


    def postorder_traverse(self):
        """
        The `postorder_traverse()` method returns a list of all keys
        in the tree in post-order order
        """

        keys = []

        if not self.node:
            return keys
        
        keys.extend(self.node.left.postorder_traverse())
        keys.extend(self.node.right.postorder_traverse())
        keys.append(self.node.vp.index)

        return keys


    def __next_gte_node(self, node, key):
        # Gets the first node with a key greater than provided key
        if node == None:
            return None
        
        if key < node.key:
            if node.left:
                after = self.__next_gte_node(node.left.node, key)
                if after == None:
                    after = node
        elif key > node.key:
            if node.right:
                after = self.__next_gte_node(node.right.node, key)           
        elif node.key == key:
            after = node

        return after


    def __rebalance(self):

        # Update the trees height and balance values
        self.__update_heights(recursive=True)
        self.__update_balances(True)

        # If balance is < -1 or > 1 then rotations are still necessary to perform
        while self.balance < -1 or self.balance > 1:
            # Left subtree is larger than right subtree so rotate to the left
            if self.balance > 1:
                if self.node.left.balance < 0:
                    self.node.left.__rotate_left()
                    self.__update_heights()
                    self.__update_balances()
                
                self.__rotate_right()
                self.__update_heights()
                self.__update_balances()
            
            # Right subtree larger than left subtree so rotate to the right
            if self.balance < -1:
                if self.node.right.balance > 0:
                    self.node.right.__rotate_right()
                    self.__update_heights()
                    self.__update_balances()
                
                self.__rotate_left()
                self.__update_heights()
                self.__update_balances()


    def __update_heights(self, recursive=True):
        # Height is max height of left or right subtrees + 1 for the root
        if self.node:
            if recursive:
                if self.node.left:
                    self.node.left.__update_heights()
                if self.node.right:
                    self.node.right.__update_heights()
            self.height = 1 + max(self.node.left.height, self.node.right.height)
        else:
            self.height = -1


    def __update_balances(self, recursive=True):
        # Calculate the balance factor of the tree
        # Balance factor calculated as follows:
        #     BF = height(left_subtree) - height(right_subtree)
        if self.node:
            if recursive:
                if self.node.left:
                    self.node.left.__update_balances()
                if self.node.right:
                    self.node.right.__update_balances()
            self.balance = self.node.left.height - self.node.right.height
        else:
            self.balance = 0


    def __rotate_right(self):
        # Set self as the subtree of the left subtree

        new_root = self.node.left.node
        new_left_sub = new_root.right.node
        old_root = self.node

        self.node = new_root
        old_root.left.node = new_left_sub
        new_root.right.node = old_root

 
    def __rotate_left(self):
        # Set self as the left subtree of the right subtree

        new_root = self.node.right.node
        new_left_sub =new_root.left.node
        old_root = self.node

        self.node = new_root
        old_root.right.node = new_left_sub
        new_root.left.node = old_root



import logging
import requests
import json
import time
import threading
from decimal import Decimal
from GhostDB.ring import Ring 

logging.basicConfig(filename="ghostdb.log")

class Cache():
    """
    The `Cache` object that provides an interface between GhostDB nodes

    You must supply a `node_file`. This is the file that contains the
    IP addresses of all servers in your fleet. You can update this file
    as needed. Any changes to this file will automatically be picked up
    by the `Cache` object.

    If `http` is not supplied then `http` defaults to `True`.
    If your fleet requires communication over `HTTPS`, then pass `False`.

    If you do not supply the `port` your GhostDB fleet is configured
    to, then it defaults to the standard port of `7991`.
    """

    _DEAD_SERVERS = []
    _REVIVE_WAIT = 30

    _API_ENDPOINT_MAP = {
        "ping": "/ping",
        "put": "/put",
        "get": "/get",
        "add": "/add",
        "delete": "/delete",
        "flush": "/flush",
        "getSnitchMetrics": "/getSnitchMetrics"
    }

    class GhostNoMoreServersError(Exception):
        """
        The error thrown when no servers in your `node_file` are
        reachable.
        """

        def __init__(self):
            Exception.__init__(
                self, 
                "[ERROR]: All nodes marked as dead: Failed to establish a connection to any servers: (Check your fleet status)"
                )
    
    class GhostKeyError(Exception):
        """
        The error thrown when they key passed to a method is not a
        valid string.
        """

        def __init__(self):
            Exception.__init__(
                self,
                "[ERROR]: Key must be of type string"
            )

    def __init__(self, nodes_file, http=True, port=7991):
        self.nodes_file = nodes_file
        self.protocol = "http://" if http else "https://"
        self.port = str(port)
        self.ring = Ring(nodes_file)
        self._server_revive()


    def get(self, key):
        """
        The `get()` method will fetch a value for a given key.

        The `key` parameter is a string
        """

        if not isinstance(key, str):
            raise Cache.GhostKeyError()
        else:
            request_object = {
                "Key": key
            }

            node = self.ring.get_point_for(key)
            
            if node == None:
                raise Cache.GhostNoMoreServersError()
                
            try:
                address = self.protocol + node.node + ":" + self.port + Cache._API_ENDPOINT_MAP["get"]
                response = requests.post(address, json=request_object)
                return json.loads(response.text)
            except:
                Cache._DEAD_SERVERS.append(node.node)
                self.ring.delete(node.node)
                return self.get(key)


    def put(self, key, value, ttl=-1):
        """
        The `put()` method will add a key/value pair into the 
        cache. If the `key` already exists in the cache, it's value
        will be overwritten.

        The `key` parameter must be a string

        The `value` parameter can be of whatever type you want as 
        long as it is JSON serializable.

        If the key/value pair is successully added, the `STORED`
        message will be returned. Otherwise the `NOT_STORED` message
        will be returned.
        """

        if not isinstance(key, str):
            raise Cache.GhostKeyError()
        else:
            request_object = {
                "Key": key,
                "Value": json.dumps(value, default=self._default),
                "TTL": ttl
            }

            node = self.ring.get_point_for(key)

            if node == None:
                raise Cache.GhostNoMoreServersError()

            try:
                address = self.protocol + node.node + ":" + self.port + Cache._API_ENDPOINT_MAP["put"]
                response = requests.post(address, json=request_object)
                return json.loads(response.text)
            except:
                Cache._DEAD_SERVERS.append(node.node)
                self.ring.delete(node.node)
                return self.put(key, value)

    def add(self, key, value, ttl=-1):
        """
        The `add()` method will add a key/value pair into the
        cache only if the `key` does not pre-exist in the cache.

        The `key` parameter must be a string

        The `value` parameter can be of whatever type you want
        as long as it is JSON serializable

        If the key/value pair is successully added, the `STORED`
        message will be returned. Otherwise the `NOT_STORED` message
        will be returned.
        """

        if not isinstance(key, str):
            raise Cache.GhostKeyError()
        else:
            request_object = {
                "Key": key,
                "Value": json.dumps(value, default=self._default),
                "TTL": ttl
            }

            node = self.ring.get_point_for(key)

            if node == None:
                raise Cache.GhostNoMoreServersError()

            try:
                address = self.protocol + node.node + ":" + self.port + Cache._API_ENDPOINT_MAP["add"]
                response = requests.post(address, json=request_object)
                return json.loads(response.text)
            except:
                Cache._DEAD_SERVERS.append(node.node)
                self.ring.delete(node.node)
                return self.add(key, value)

    def delete(self, key):
        """
        The `delete()` method will remove a key/value pair
        from the cache specified by the `key` parameter

        The `key` parameter must be a string

        If the key/value pair is successfully deleted, 
        the `REMOVED` message will be returned, otherwise
        the `NOT_FOUND` message will be returned
        """
        
        if not isinstance(key, str):
            raise Cache.GhostKeyError()
        else:
            request_object = {
                "Key": key
            }

            node = self.ring.get_point_for(key)

            if node == None:
                raise Cache.GhostNoMoreServersError()

            try:
                address = self.protocol + node.node + ":" + self.port + Cache._API_ENDPOINT_MAP["delete"]
                response = requests.post(address, json=request_object)
                return json.loads(response.text)
            except:
                Cache._DEAD_SERVERS.append(node.node)
                self.ring.delete(node.node)
                return self.delete(node.node)

    def flush(self):
        """
        The `flush()` method will delete all key/value
        pairs from all nodes specified in the `node_file`
        at the time of flushing.
        """

        nodes = self.ring.get_points()
        
        for node in nodes:
            try:
                address = self.protocol + node.node + ":" + self.port + Cache._API_ENDPOINT_MAP["flush"]
                response = requests.get(address)
                return json.loads(response.text)
            except:
                Cache._DEAD_SERVERS.append(node.node)
                self.ring.delete(node.node)
                return self.flush()

    def getSnitchMetrics(self, metrics=None, visitedNodes=None):
        """
        The `getSnitchMetrics()` method will fetch all
        snitch metricsfrom all nodes specified in the
        `node_file` at the time of calling.
        """
        if metrics is None:
            metrics = []
        
        if visitedNodes is None:
            visitedNodes = []

        nodes = self.ring.get_points()
        
        for node in nodes:
            try:
                if node.node not in visitedNodes:
                    address = self.protocol + node.node + ":" + self.port + Cache._API_ENDPOINT_MAP["getSnitchMetrics"]
                    response = requests.get(address)
                    nodeMetrics = json.loads(response.text)
                    metrics.append({"node": node.node, "metrics": nodeMetrics})
                    visitedNodes.append(node.node)
            except:
                Cache._DEAD_SERVERS.append(node.node)
                self.ring.delete(node.node)
                return self.getSnitchMetrics(metrics, visitedNodes)
        return metrics

    def _server_revive(self):
        for i, server in enumerate(Cache._DEAD_SERVERS):
            try:
                address = self.protocol + server + ":" + self.port + Cache._API_ENDPOINT_MAP["ping"]
                response = requests.get(address)
                if response.status_code == 200:
                    self.ring.add(server)
                    Cache._DEAD_SERVERS.pop(i)
            except Exception as e:
                continue
        threading.Timer(Cache._REVIVE_WAIT, self._server_revive).start()

    def _default(obj):
        if isinstance(obj, Decimal):
            return str(obj)
        raise TypeError("Object of type '%s' is not JSON serializable" % type(obj).__name__)


from GhostDB.avl_tree import AVLTree
from GhostDB.virtual_point import VirtualPoint
import binascii

class Ring():
    """
    The `Ring` object represents the consistent hashing ring.

    The `node_config` parameter defaults to `None`, however it
    should be provided by the `Cache` object

    The `replicas` parameter adds replicas for a virtual point
    in the tree. This defaults to `1` and it is not recommended
    to change this.
    """

    def __init__(self, node_config=None, replicas=1):
        self.replicas = replicas
        self.ring = AVLTree()
   
        if node_config:
            with open(node_config, 'r') as nodes:
                for node in nodes.readlines():
                    node = node.strip()
                    self.add(node)

    def __len__(self):
        return len(self.ring.get_nodes())


    def __str__(self):
        return self.ring.__str__()


    def add(self, node):
        """
        The `add()` method adds a GhostDB node to the
        consistent hashing ring.
        """

        for i in range(self.replicas):
            key = self.key_hash(node, i)
            vp = VirtualPoint(node, key)
            self.ring.insert(key, vp)
    

    def delete(self, node):
        """
        The `delete()` method removes a GhostDB node from the
        consistent hashing ring. This is typically performed
        if the GhostDB node is unreachable.
        """

        for i in range(self.replicas):
            key = self.key_hash(node, i)
            self.ring.remove(key)


    def get_point_for(self, key):
        """
        The `get_point_for()` method returns the correct 
        GhostDB node to send a request to for a given key.
        """

        if len(self) == 0:
            return None
        key = self.key_hash(key)
        node_key, node_value = self.ring.next_gte_pair(key)
        if not node_value:
            node_key, node_value = self.ring.minimum_pair()

        return node_value


    def get_points(self):
        """
        The `get_points()` method returns all GhostDB nodes
        in the consistent hashing ring.
        """
        
        return self.ring.get_nodes()

    def key_hash(self, key, index=None):
        """
        The `key_hash()` method generates and returns 
        the unsigned CRC32 hash for a provided key in 
        hexidecimal form.
        """

        if index:
            key = "{:s}:{:d}".format(key, index)
        s = binascii.crc32(bytes(key, 'utf-8'))
        return hex(s)



class VirtualPoint():
    """
    Virtual Point represents a point of a GhostDB cache on the hash ring
    """
    def __init__(self, node, index):
        self.node = node
        self.index = index

    def __str__(self):
        return "====\nNode: {:s}\nIndex: {:s}\n====".format(self.node, self.index) 

    def set_index(index):
        self.index = index



from setuptools import setup

def readme():
    with open("README.rst") as f:
        return f.read()

setup(
    name="ghostdb",
    version="0.3",
    description="Python SDK for GhostDB Cache",
    long_description=readme(),
    long_description_content_type="text/markdown",
    author="Jake Grogan",
    author_email="jake.grogan8@mail.dcu.ie",
    license="MIT",
    packages=["GhostDB"],
    package_data={
        "GhostDB": [
            "tests/ghostdb_sim.conf",
            "tests/ghostdb_test.conf"
        ]
    },
    install_requires=[
        "requests",
    ],
    zip_safe=False,
    test_suite='nose.collector',
    tests_require=['nose'],
)